import logging
from enum import Enum
from functools import partial
from os import PathLike
from pathlib import Path
from typing import Generator, Optional, Sequence, Tuple, Union

import numpy as np
import pandas as pd
from scipy.ndimage import distance_transform_edt
from scipy.spatial.distance import pdist, squareform
from skimage.measure import regionprops

from .. import io
from ._measurement import SteinbockMeasurementException

logger = logging.getLogger(__name__)


class SteinbockNeighborsMeasurementException(SteinbockMeasurementException):
    pass


def _expand_mask_euclidean(mask: np.ndarray, dmax: float) -> np.ndarray:
    dists, (i, j) = distance_transform_edt(mask == 0, return_indices=True)
    return np.where(dists <= dmax, mask[i, j], mask)


def _to_triu_indices(k: np.ndarray, n: int) -> Tuple[np.ndarray, np.ndarray]:
    # https://stackoverflow.com/a/27088560 (optimized for numpy arrays)
    # Briefly, the formulas were derived using triangular numbers/roots
    i = n - 2 - np.floor((-8 * k + 4 * n * (n - 1) - 7) ** 0.5 / 2 - 0.5)
    j = k + i + 1 - (n * (n - 1) - (n - i) * (n - i - 1)) // 2
    return i.astype(int), j.astype(int)


def _measure_centroid_distance_neighbors(
    mask: np.ndarray,
    metric: Optional[str] = None,
    dmax: Optional[float] = None,
    kmax: Optional[int] = None,
) -> pd.DataFrame:
    if metric is None:
        raise SteinbockNeighborsMeasurementException("Metric is required")
    props = regionprops(mask)
    labels = np.array([p.label for p in props])
    centroids = np.array([p.centroid for p in props])
    condensed_dists = pdist(centroids, metric=metric)
    if kmax is not None:
        k = min(kmax, len(props) - 1)
        dist_mat = squareform(condensed_dists, checks=False).astype(float)
        np.fill_diagonal(dist_mat, np.inf)
        knn_indices = np.argpartition(dist_mat, k - 1)[:, :k]
        if dmax is not None:
            knn_dists = np.take_along_axis(dist_mat, knn_indices, -1)
            indices1, indices2 = np.nonzero(knn_dists <= dmax)
            indices2 = knn_indices[(indices1, indices2)]
        else:
            indices1 = np.repeat(np.arange(len(props)), k)
            indices2 = np.ravel(knn_indices)
        distances = dist_mat[(indices1, indices2)]
    elif dmax is not None:
        (condensed_indices,) = np.nonzero(condensed_dists <= dmax)
        indices1, indices2 = _to_triu_indices(condensed_indices, len(props))
        distances = condensed_dists[condensed_indices]
        indices1, indices2, distances = (
            np.concatenate((indices1, indices2)),
            np.concatenate((indices2, indices1)),
            np.concatenate((distances, distances)),
        )
    else:
        raise SteinbockNeighborsMeasurementException(
            "Specify either dmax or kmax (or both)"
        )
    return pd.DataFrame(
        data={
            "Object": np.asarray(labels[indices1], dtype=io.mask_dtype),
            "Neighbor": np.asarray(labels[indices2], dtype=io.mask_dtype),
            "Distance": np.asarray(distances, dtype=np.float32),
        }
    )


def _measure_euclidean_border_distance_neighbors(
    mask: np.ndarray,
    metric: Optional[str] = None,
    dmax: Optional[float] = None,
    kmax: Optional[int] = None,
) -> pd.DataFrame:
    if metric not in (None, "euclidean"):
        raise SteinbockNeighborsMeasurementException("Metric has to be euclidean")
    labels1 = []
    labels2 = []
    distances = []
    unique_labels = np.unique(mask)
    unique_labels = unique_labels[unique_labels != 0]
    for label in unique_labels:
        if dmax is not None:
            ys, xs = np.nonzero(mask == label)
            dmax_int = int(np.ceil(dmax))
            ymin, ymax = np.amin(ys) - dmax_int, np.amax(ys) + dmax_int
            xmin, xmax = np.amin(xs) - dmax_int, np.amax(xs) + dmax_int
            mask_or_patch = mask[
                max(0, ymin) : min(mask.shape[0], ymax + 1),
                max(0, xmin) : min(mask.shape[1], xmax + 1),
            ]
            neighbor_labels = np.unique(mask_or_patch)
            neighbor_labels = neighbor_labels[neighbor_labels != 0]
        else:
            mask_or_patch = mask
            neighbor_labels = unique_labels
        dists = distance_transform_edt(mask_or_patch != label)
        neighbor_labels = neighbor_labels[neighbor_labels != label]
        neighbor_dists = np.array(
            [
                np.amin(dists[mask_or_patch == neighbor_label])
                for neighbor_label in neighbor_labels
            ]
        )
        if dmax is not None:
            neighbor_labels = neighbor_labels[neighbor_dists <= dmax]
            neighbor_dists = neighbor_dists[neighbor_dists <= dmax]
        if kmax is not None and len(neighbor_labels) > kmax:
            knn_indices = np.argpartition(neighbor_dists, kmax - 1)[:kmax]
            neighbor_labels = neighbor_labels[knn_indices]
            neighbor_dists = neighbor_dists[knn_indices]
        labels1 += [label] * len(neighbor_labels)
        labels2 += neighbor_labels.tolist()
        distances += neighbor_dists.tolist()
    return pd.DataFrame(
        data={
            "Object": np.asarray(labels1, dtype=io.mask_dtype),
            "Neighbor": np.asarray(labels2, dtype=io.mask_dtype),
            "Distance": np.asarray(distances, dtype=np.float32),
        }
    )


def _measure_euclidean_pixel_expansion_neighbors(
    mask: np.ndarray,
    metric: Optional[str] = None,
    dmax: Optional[float] = None,
    kmax: Optional[int] = None,
) -> pd.DataFrame:
    if metric not in (None, "euclidean"):
        raise SteinbockNeighborsMeasurementException(
            "Metric has to be Euclidean for pixel expansion"
        )
    if dmax is None:
        raise SteinbockNeighborsMeasurementException(
            "Maximum distance required for pixel expansion"
        )
    if kmax is not None:
        raise SteinbockNeighborsMeasurementException(
            "k-nearest neighbors is not supported by pixel expansion"
        )
    mask = _expand_mask_euclidean(mask, dmax)
    neighbors = _measure_euclidean_border_distance_neighbors(mask, dmax=1.0)
    neighbors["Distance"] = np.nan
    return neighbors


class NeighborhoodType(Enum):
    """"""

    CENTROID_DISTANCE = partial(_measure_centroid_distance_neighbors)
    EUCLIDEAN_BORDER_DISTANCE = partial(_measure_euclidean_border_distance_neighbors)
    EUCLIDEAN_PIXEL_EXPANSION = partial(_measure_euclidean_pixel_expansion_neighbors)


def measure_neighbors(
    mask: np.ndarray,
    neighborhood_type: NeighborhoodType,
    metric: Optional[str] = None,
    dmax: Optional[float] = None,
    kmax: Optional[int] = None,
) -> pd.DataFrame:
    return neighborhood_type.value(mask, metric=metric, dmax=dmax, kmax=kmax)


def try_measure_neighbors_from_disk(
    mask_files: Sequence[Union[str, PathLike]],
    neighborhood_type: NeighborhoodType,
    metric: Optional[str] = None,
    dmax: Optional[float] = None,
    kmax: Optional[int] = None,
    mmap: bool = False,
) -> Generator[Tuple[Path, pd.DataFrame], None, None]:
    for mask_file in mask_files:
        try:
            if mmap:
                mask = io.mmap_mask(mask_file)
            else:
                mask = io.read_mask(mask_file)
            neighbors = measure_neighbors(
                mask, neighborhood_type, metric=metric, dmax=dmax, kmax=kmax
            )
            del mask
            yield Path(mask_file), neighbors
            del neighbors
        except:
            logger.exception(f"Error measuring neighbors in {mask_file}")



import sys
from pathlib import Path

import click
import click_log

from .._env import check_x11, run_captured, use_ilastik_env
from .._steinbock import SteinbockException
from .._steinbock import logger as steinbock_logger
from .utils import OrderedClickGroup, catch_exception


@click.group(name="apps", cls=OrderedClickGroup, help="Third-party applications")
def apps_cmd_group():
    pass


@apps_cmd_group.command(
    name="ilastik",
    context_settings={"ignore_unknown_options": True},
    help="Run Ilastik (GUI requires X11)",
    add_help_option=False,
)
@click.option(
    "--ilastik",
    "ilastik_binary",
    type=click.STRING,
    default="/opt/ilastik/run_ilastik.sh",
    show_default=True,
    help="Ilastik binary",
)
@click.argument("ilastik_args", nargs=-1, type=click.UNPROCESSED)
@click_log.simple_verbosity_option(logger=steinbock_logger)
@catch_exception(handle=SteinbockException)
@check_x11
@use_ilastik_env
def ilastik_cmd(ilastik_binary, ilastik_args, ilastik_env):
    args = [ilastik_binary] + list(ilastik_args)
    result = run_captured(args, env=ilastik_env)
    sys.exit(result.returncode)


@apps_cmd_group.command(
    name="cellprofiler",
    context_settings={"ignore_unknown_options": True},
    help="Run CellProfiler (GUI requires X11)",
    add_help_option=False,
)
@click.option(
    "--cellprofiler",
    "cellprofiler_binary",
    type=click.STRING,
    default="cellprofiler",
    show_default=True,
    help="CellProfiler binary",
)
@click.option(
    "--plugins-directory",
    "cellprofiler_plugin_dir",
    type=click.Path(file_okay=False),
    default="/opt/cellprofiler_plugins",
    show_default=True,
    help="Path to the CellProfiler plugin directory",
)
@click.argument("cellprofiler_args", nargs=-1, type=click.UNPROCESSED)
@click_log.simple_verbosity_option(logger=steinbock_logger)
@catch_exception(handle=SteinbockException)
@check_x11
def cellprofiler_cmd(cellprofiler_binary, cellprofiler_plugin_dir, cellprofiler_args):
    args = [cellprofiler_binary] + list(cellprofiler_args)
    if Path(cellprofiler_plugin_dir).exists():
        args.append(f"--plugins-directory={cellprofiler_plugin_dir}")
    result = run_captured(args)
    sys.exit(result.returncode)



import click

from ..._cli.utils import OrderedClickGroup
from .ilastik import ilastik_cmd_group


@click.group(
    name="classify",
    cls=OrderedClickGroup,
    help="Perform pixel classification to extract probabilities",
)
def classify_cmd_group():
    pass


classify_cmd_group.add_command(ilastik_cmd_group)



import shutil
from pathlib import Path

import pytest
from steinbock.segmentation import cellprofiler

cellprofiler_binary = "cellprofiler"
cellprofiler_plugin_dir = "/opt/cellprofiler_plugins"


class TestCellprofilerSegmentation:
    def test_create_and_save_segmentation_pipeline(self, tmp_path: Path):
        cellprofiler.create_and_save_segmentation_pipeline(
            tmp_path / "cell_segmentation.cppipe"
        )  # TODO

    @pytest.mark.skip(reason="Test would take too long")
    @pytest.mark.skipif(
        shutil.which(cellprofiler_binary) is None,
        reason="CellProfiler is not available",
    )
    def test_try_segment_objects(
        self, imc_test_data_steinbock_path: Path, tmp_path: Path
    ):
        cellprofiler.try_segment_objects(
            cellprofiler_binary,
            imc_test_data_steinbock_path / "cell_segmentation.cppipe",
            imc_test_data_steinbock_path / "ilastik_probabilities",
            tmp_path / "masks",
            cellprofiler_plugin_dir=cellprofiler_plugin_dir,
        )  # TODO



import subprocess
from importlib import resources
from os import PathLike
from pathlib import Path
from typing import Union

from ..._env import run_captured
from . import data as cellprofiler_data


def create_and_save_measurement_pipeline(
    measurement_pipeline_file: Union[str, PathLike], num_channels: int
) -> None:
    s = resources.read_text(cellprofiler_data, "cell_measurement.cppipe")
    s = s.replace("{{NUM_CHANNELS}}", str(num_channels))
    with Path(measurement_pipeline_file).open(mode="w") as f:
        f.write(s)


def try_measure_objects(
    cellprofiler_binary: str,
    measurement_pipeline_file: Union[str, PathLike],
    cpdata_dir: Union[str, PathLike],
    cpout_dir: Union[str, PathLike],
    cellprofiler_plugin_dir: Union[str, PathLike, None] = None,
) -> subprocess.CompletedProcess:
    args = [
        cellprofiler_binary,
        "-c",
        "-r",
        "-p",
        str(measurement_pipeline_file),
        "-i",
        str(cpdata_dir),
        "-o",
        str(cpout_dir),
    ]
    if cellprofiler_plugin_dir is not None and Path(cellprofiler_plugin_dir).exists():
        args.append(f"--plugins-directory={cellprofiler_plugin_dir}")
    return run_captured(args)



from pathlib import Path

import click
import click_log

from ... import io
from ..._cli.utils import catch_exception, logger
from ..._steinbock import SteinbockException
from ..._steinbock import logger as steinbock_logger
from ..regionprops import try_measure_regionprops_from_disk


@click.command(name="regionprops", help="Measure object region properties")
@click.option(
    "--img",
    "img_dir",
    type=click.Path(exists=True, file_okay=False),
    default="img",
    show_default=True,
    help="Path to the image directory",
)
@click.option(
    "--masks",
    "mask_dir",
    type=click.Path(exists=True, file_okay=False),
    default="masks",
    show_default=True,
    help="Path to the mask directory",
)
@click.option(
    "--mmap/--no-mmap",
    "mmap",
    default=False,
    show_default=True,
    help="Use memory mapping for reading images/masks",
)
@click.argument("skimage_regionprops", nargs=-1, type=click.STRING)
@click.option(
    "-o",
    "regionprops_dir",
    type=click.Path(file_okay=False),
    default="regionprops",
    show_default=True,
    help="Path to the object region properties output directory",
)
@click_log.simple_verbosity_option(logger=steinbock_logger)
@catch_exception(handle=SteinbockException)
def regionprops_cmd(img_dir, mask_dir, mmap, skimage_regionprops, regionprops_dir):
    img_files = io.list_image_files(img_dir)
    mask_files = io.list_mask_files(mask_dir, base_files=img_files)
    Path(regionprops_dir).mkdir(exist_ok=True)
    if not skimage_regionprops:
        skimage_regionprops = [
            "area",
            "centroid",
            "major_axis_length",
            "minor_axis_length",
            "eccentricity",
        ]
    for img_file, mask_file, regionprops in try_measure_regionprops_from_disk(
        img_files, mask_files, skimage_regionprops, mmap=mmap
    ):
        regionprops_file = io._as_path_with_suffix(
            Path(regionprops_dir) / img_file.name, ".csv"
        )
        io.write_data(regionprops, regionprops_file)
        logger.info(regionprops_file)
        del regionprops



import logging
from collections import Counter
from os import PathLike
from pathlib import Path
from typing import Generator, Sequence, Tuple, Union

import networkx as nx
import pandas as pd

from .. import io

logger = logging.getLogger(__name__)


def convert_to_networkx(neighbors: pd.DataFrame, *data_list) -> nx.Graph:
    edges = neighbors[["Object", "Neighbor"]].astype(int).values.tolist()
    undirected_edges = [tuple(sorted(edge)) for edge in edges]
    is_directed = any([x != 2 for x in Counter(undirected_edges).values()])
    graph: nx.Graph = nx.from_pandas_edgelist(
        neighbors,
        source="Object",
        target="Neighbor",
        edge_attr=True,
        create_using=nx.DiGraph if is_directed else nx.Graph,
    )
    if len(data_list) > 0:
        merged_data = data_list[0]
        for data in data_list[1:]:
            merged_data = pd.merge(merged_data, data, left_index=True, right_index=True)
        node_attributes = {
            int(object_id): object_data.to_dict()
            for object_id, object_data in merged_data.iterrows()
        }
        nx.set_node_attributes(graph, node_attributes)
    return graph


def try_convert_to_networkx_from_disk(
    neighbors_files: Sequence[Union[str, PathLike]], *data_file_lists
) -> Generator[Tuple[Path, Tuple[Path, ...], nx.Graph], None, None]:
    for neighbors_file, *data_files in zip(neighbors_files, *data_file_lists):
        data_files = tuple(Path(data_file) for data_file in data_files)
        try:
            neighbors = io.read_neighbors(neighbors_file)
            data_list = [io.read_data(data_file) for data_file in data_files]
            graph = convert_to_networkx(neighbors, *data_list)
            yield Path(neighbors_file), data_files, graph
            del neighbors, data_list, graph
        except:
            logger.exception(f"Error converting {neighbors_file} to networkx")



from pathlib import Path

import numpy as np
from steinbock import io
from steinbock.measurement import intensities
from steinbock.measurement.intensities import IntensityAggregation


class TestIntensitiesMeasurement:
    def test_measure_intensites(self):
        img = np.array(
            [
                [
                    [0.5, 1.5, 0.1],
                    [0.1, 0.2, 0.3],
                    [0.1, 6.5, 3.5],
                ],
                [
                    [20, 30, 1],
                    [1, 2, 3],
                    [1, 100, 200],
                ],
            ],
            dtype=io.img_dtype,
        )
        mask = np.array(
            [
                [1, 1, 0],
                [0, 0, 0],
                [0, 2, 2],
            ],
            dtype=io.mask_dtype,
        )
        channel_names = ["Channel 1", "Channel 2"]
        df = intensities.measure_intensites(
            img, mask, channel_names, IntensityAggregation.MEAN
        )
        assert np.all(df.index.values == np.array([1, 2]))
        assert np.all(df.columns.values == np.array(channel_names))
        assert np.all(df.values == np.array([[1.0, 25.0], [5.0, 150.0]]))

    def test_try_measure_intensities_from_disk(
        self, imc_test_data_steinbock_path: Path
    ):
        panel = io.read_panel(imc_test_data_steinbock_path / "panel.csv")
        img_files = io.list_image_files(imc_test_data_steinbock_path / "img")
        mask_files = io.list_mask_files(
            imc_test_data_steinbock_path / "masks", base_files=img_files
        )
        gen = intensities.try_measure_intensities_from_disk(
            img_files,
            mask_files,
            panel["name"].tolist(),
            IntensityAggregation.MEAN,
        )
        for img_file, mask_file, df in gen:
            pass  # TODO



from pathlib import Path

import numpy as np
import pytest
from steinbock import io
from steinbock.preprocessing import imc


@pytest.mark.skipif(not imc.imc_available, reason="IMC is not available")
class TestIMCPreprocessing:
    def test_list_mcd_files(self, imc_test_data_steinbock_path: Path):
        imc.list_mcd_files(imc_test_data_steinbock_path / "raw")  # TODO

    def test_list_txt_files(self, imc_test_data_steinbock_path: Path):
        imc.list_txt_files(imc_test_data_steinbock_path / "raw")  # TODO

    def test_filter_hot_pixels(self):
        img = np.array(
            [
                [
                    [1, 2, 3],
                    [4, 13, 6],
                    [7, 8, 9],
                ]
            ],
            dtype=io.img_dtype,
        )
        filtered_img = imc.filter_hot_pixels(img, 3.0)
        expected_filtered_img = np.array(
            [
                [
                    [1, 2, 3],
                    [4, 9, 6],
                    [7, 8, 9],
                ]
            ],
            dtype=io.img_dtype,
        )
        assert np.all(filtered_img == expected_filtered_img)

    def test_preprocess_image(self):
        img = np.array(
            [
                [
                    [1, 2, 3],
                    [4, 13, 6],
                    [7, 8, 9],
                ],
            ],
            dtype=io.img_dtype,
        )
        preprocessed_img = imc.preprocess_image(img, hpf=3.0)
        expected_preprocessed_img = np.array(
            [
                [
                    [1, 2, 3],
                    [4, 9, 6],
                    [7, 8, 9],
                ]
            ],
            dtype=io.img_dtype,
        )
        assert np.all(preprocessed_img == expected_preprocessed_img)

    def test_try_preprocess_images_from_disk(self, imc_test_data_steinbock_path: Path):
        mcd_files = imc.list_mcd_files(imc_test_data_steinbock_path / "raw")
        txt_files = imc.list_txt_files(imc_test_data_steinbock_path / "raw")
        gen = imc.try_preprocess_images_from_disk(mcd_files, txt_files)
        for mcd_txt_file, acquisition, img, recovery_file, recovered in gen:
            pass  # TODO

    def test_create_panel_from_imc_panel(self, imc_test_data_steinbock_path: Path):
        imc.create_panel_from_imc_panel(
            imc_test_data_steinbock_path / "raw" / "panel.csv"
        )  # TODO

    def test_create_panel_from_mcd_file(self, imc_test_data_steinbock_path: Path):
        mcd_files = imc.list_mcd_files(
            imc_test_data_steinbock_path / "raw" / "20210305_NE_mockData1"
        )
        imc.create_panel_from_mcd_files(mcd_files)  # TODO

    def test_create_panel_from_txt_file(self, imc_test_data_steinbock_path: Path):
        txt_files = imc.list_txt_files(
            imc_test_data_steinbock_path / "raw" / "20210305_NE_mockData1"
        )
        imc.create_panel_from_txt_files(txt_files)  # TODO



import shutil
import sys
from pathlib import Path

import click
import click_log
import numpy as np

from ... import io
from ..._cli.utils import (
    OrderedClickGroup,
    SteinbockCLIException,
    catch_exception,
    logger,
)
from ..._env import use_ilastik_env
from ..._steinbock import SteinbockException
from ..._steinbock import logger as steinbock_logger
from .. import ilastik


@click.group(
    name="ilastik",
    cls=OrderedClickGroup,
    help="Perform supervised pixel classification using Ilastik",
)
def ilastik_cmd_group():
    pass


@ilastik_cmd_group.command(
    name="prepare", help="Prepare an Ilastik project file, images and crops"
)
@click.option(
    "-o",
    "ilastik_project_file",
    type=click.Path(dir_okay=False),
    default="pixel_classifier.ilp",
    show_default=True,
    help="Path to the Ilastik project output file",
)
@click.option(
    "--img",
    "img_dir",
    type=click.Path(exists=True, file_okay=False),
    default="img",
    show_default=True,
    help="Path to the image directory",
)
@click.option(
    "--panel",
    "panel_file",
    type=click.Path(exists=True, dir_okay=False),
    default="panel.csv",
    show_default=True,
    help="Path to the panel file",
)
@click.option(
    "--aggr",
    "aggr_func_name",
    type=click.STRING,
    default="mean",
    show_default=True,
    help="Numpy function for aggregating channel pixels",
)
@click.option(
    "--mean/--no-mean",
    "prepend_mean",
    default=True,
    show_default=True,
    help="Prepend mean of all channels as a new channel",
)
@click.option(
    "--meanfactor",
    "mean_factor",
    type=click.FLOAT,
    default=100.0,
    show_default=True,
    help="Factor to multiply the channels mean with",
)
@click.option(
    "--scale",
    "scale_factor",
    type=click.INT,
    default=2,
    show_default=True,
    help="Ilastik image scale factor",
)
@click.option(
    "--cropsize",
    "ilastik_crop_size",
    type=click.INT,
    default=512,
    show_default=True,
    help="Ilastik crop size (in pixels)",
)
@click.option(
    "--imgout",
    "ilastik_img_dir",
    type=click.Path(file_okay=False),
    default="ilastik_img",
    show_default=True,
    help="Path to the Ilastik image output directory",
)
@click.option(
    "--cropout",
    "ilastik_crop_dir",
    type=click.Path(file_okay=False),
    default="ilastik_crops",
    show_default=True,
    help="Path to the Ilastik crop output directory",
)
@click.option("--seed", "seed", type=click.INT, help="Random seed")
@click_log.simple_verbosity_option(logger=steinbock_logger)
@catch_exception(handle=SteinbockException)
def prepare_cmd(
    ilastik_project_file,
    img_dir,
    panel_file,
    aggr_func_name,
    prepend_mean,
    mean_factor,
    scale_factor,
    ilastik_crop_size,
    ilastik_img_dir,
    ilastik_crop_dir,
    seed,
):
    channel_groups = None
    if Path(panel_file).exists():
        panel = io.read_panel(panel_file)
        if "ilastik" in panel and panel["ilastik"].notna().any():
            channel_groups = panel["ilastik"].values
    aggr_func = getattr(np, aggr_func_name)
    img_files = io.list_image_files(img_dir)
    Path(ilastik_img_dir).mkdir(exist_ok=True)
    ilastik_img_files = []
    for img_file, ilastik_img in ilastik.try_create_ilastik_images_from_disk(
        img_files,
        channel_groups=channel_groups,
        aggr_func=aggr_func,
        prepend_mean=prepend_mean,
        mean_factor=mean_factor,
        scale_factor=scale_factor,
    ):
        ilastik_img_file = io._as_path_with_suffix(
            Path(ilastik_img_dir) / img_file.name, ".h5"
        )
        ilastik.write_ilastik_image(ilastik_img, ilastik_img_file)
        ilastik_img_files.append(ilastik_img_file)
        logger.info(ilastik_img_file)
        del ilastik_img
    Path(ilastik_crop_dir).mkdir(exist_ok=True)
    ilastik_crop_files = []
    for (
        ilastik_img_file,
        ilastik_crop_x,
        ilastik_crop_y,
        ilastik_crop,
    ) in ilastik.try_create_ilastik_crops_from_disk(
        ilastik_img_files, ilastik_crop_size, np.random.default_rng(seed=seed)
    ):
        if ilastik_crop is not None:
            ilastik_crop_file = Path(ilastik_crop_dir) / (
                f"{ilastik_img_file.stem}"
                f"_x{ilastik_crop_x}_y{ilastik_crop_y}"
                f"_w{ilastik_crop_size}_h{ilastik_crop_size}.h5"
            )
            ilastik.write_ilastik_crop(ilastik_crop, ilastik_crop_file)
            ilastik_crop_files.append(ilastik_crop_file)
            logger.info(ilastik_crop_file)
            del ilastik_crop
        else:
            logger.warning(f"Image {ilastik_img_file} too small for crop size")
    ilastik.create_and_save_ilastik_project(ilastik_crop_files, ilastik_project_file)
    logger.info(ilastik_project_file)


@ilastik_cmd_group.command(
    name="run", help="Run a pixel classification batch using Ilastik"
)
@click.option(
    "--ilastik",
    "ilastik_binary",
    type=click.STRING,
    default="/opt/ilastik/run_ilastik.sh",
    show_default=True,
    help="Ilastik binary",
)
@click.option(
    "--ilp",
    "ilastik_project_file",
    type=click.Path(exists=True, dir_okay=False),
    default="pixel_classifier.ilp",
    show_default=True,
    help="Path to the Ilastik project file",
)
@click.option(
    "--img",
    "ilastik_img_dir",
    type=click.Path(exists=True, file_okay=False),
    default="ilastik_img",
    show_default=True,
    help="Path to the Ilastik image directory",
)
@click.option(
    "-o",
    "ilastik_probab_dir",
    type=click.Path(file_okay=False),
    default="ilastik_probabilities",
    show_default=True,
    help="Path to the Ilastik probabilities output directory",
)
@click.option(
    "--threads",
    "num_threads",
    type=click.IntRange(min=0),
    help="Maximum number of threads to use",
)
@click.option(
    "--mem",
    "memory_limit",
    type=click.IntRange(min=0),
    help="Memory limit (in megabytes)",
)
@click_log.simple_verbosity_option(logger=steinbock_logger)
@catch_exception(handle=SteinbockException)
@use_ilastik_env
def run_cmd(
    ilastik_binary,
    ilastik_project_file,
    ilastik_img_dir,
    ilastik_probab_dir,
    num_threads,
    memory_limit,
    ilastik_env,
):
    ilastik_img_files = ilastik.list_ilastik_image_files(ilastik_img_dir)
    Path(ilastik_probab_dir).mkdir(exist_ok=True)
    result = ilastik.run_pixel_classification(
        ilastik_binary,
        ilastik_project_file,
        ilastik_img_files,
        ilastik_probab_dir,
        num_threads=num_threads,
        memory_limit=memory_limit,
        ilastik_env=ilastik_env,
    )
    sys.exit(result.returncode)


@ilastik_cmd_group.command(
    name="fix", help="Fix existing Ilastik training data (experimental)"
)
@click.option(
    "--ilp",
    "ilastik_project_file",
    type=click.Path(exists=True, dir_okay=False),
    default="pixel_classifier.ilp",
    show_default=True,
    help="Path to the Ilastik project file",
)
@click.option(
    "--crop",
    "ilastik_crop_dir",
    type=click.Path(exists=True, file_okay=False),
    default="ilastik_crops",
    show_default=True,
    help="Path to the Ilastik crop directory",
)
@click.option(
    "--probabs",
    "ilastik_probab_dir",
    type=click.Path(file_okay=False),
    default="ilastik_probabilities",
    show_default=True,
    help="Path to the Ilastik probabilities directory",
)
@click.option(
    "--backup/--no-backup",
    "create_backup",
    default=True,
    show_default=True,
    help="Backup files before attempting to patch them",
)
@click.option(
    "--axisorder",
    "orig_axis_order",
    type=click.STRING,
    help="Axis order of the existing crops (e.g. zyxc)",
)
@click_log.simple_verbosity_option(logger=steinbock_logger)
@catch_exception(handle=SteinbockException)
def fix_cmd(
    ilastik_project_file,
    ilastik_crop_dir,
    ilastik_probab_dir,
    create_backup,
    orig_axis_order,
):
    ilastik_crop_files = ilastik.list_ilastik_crop_files(ilastik_crop_dir)
    if create_backup:
        ilastik_project_backup_file = Path(ilastik_project_file).with_name(
            Path(ilastik_project_file).name + ".bak"
        )
        if ilastik_project_backup_file.exists():
            raise SteinbockCLIException("Ilastik project backup file exists")
        ilastik_crop_backup_dir = Path(ilastik_crop_dir).with_name(
            Path(ilastik_crop_dir).name + ".bak"
        )
        if ilastik_crop_backup_dir.exists():
            raise SteinbockCLIException("Ilastik crop backup directory exists")
        shutil.copyfile(ilastik_project_file, ilastik_project_backup_file)
        ilastik_crop_backup_dir.mkdir()
        for ilastik_crop_file in ilastik_crop_files:
            shutil.copyfile(
                ilastik_crop_file,
                ilastik_crop_backup_dir / ilastik_crop_file.name,
            )
    ilastik_crop_shapes = {}
    last_transpose_axes = None
    for (
        ilastik_crop_file,
        transpose_axes,
        ilastik_crop,
    ) in ilastik.try_fix_ilastik_crops_from_disk(
        ilastik_crop_files, orig_axis_order=orig_axis_order
    ):
        if last_transpose_axes not in (None, transpose_axes):
            raise SteinbockCLIException("Inconsistent axis orders across crops")
        ilastik.write_ilastik_crop(ilastik_crop, ilastik_crop_file)
        ilastik_crop_shapes[ilastik_crop_file.stem] = ilastik_crop.shape
        last_transpose_axes = transpose_axes
        logger.info(ilastik_crop_file)
        del ilastik_crop
    ilastik.fix_ilastik_project_file_inplace(
        ilastik_project_file,
        ilastik_crop_dir,
        ilastik_probab_dir,
        ilastik_crop_shapes,
        last_transpose_axes,
    )
    logger.info(ilastik_project_file)
